#define axis_version  "0.9.3"  \
 
#include  "parse_types.h"
#include  "parser.tab.h"
#include  <iostream>
#include  <fstream>
#include  "buffer.h"
#include  "lexer.h"
#include  "../version.h"
#include  <csignal>
#include  "atlas-types.h"
#include  "global.h" 
#include  <cstring>
#include  <stdexcept>
#include  "axis.h"
#include  <cstdlib>


#ifdef NREADLINE
#define readline nullptr
#define add_history nullptr
#define clear_history()
#else
#include <readline/readline.h>
#include <readline/history.h>
#endif
#ifdef NOT_UNIX
#define isatty(x) true // if we cannot find out, guess it is ``yes''
#define chdir(x) (-1) // just fail for every argument
#else
#include <unistd.h>
#endif

int yylex (YYSTYPE *, YYLTYPE *);
int yyparse( atlas::interpreter::expr_p* parsed_expr, int* verbosity );

namespace { 
            const char* keywords[] =
             {"quit"
             ,"set","let","in","begin","end"
             ,"if","then","else","elif","fi"
             ,"and","or","not"
             ,"next","do","dont","from","downto","while","for","od"
             ,"case","esac", "rec_fun"
             ,"true","false", "die", "break", "return"
             ,"whattype","showall","forget"
             ,nullptr};
            
            static atlas::interpreter::shared_share input_path_pointer,prelude_log_pointer;
            #ifndef NOT_UNIX
            enum { cwd_size = 0x1000 };
            char cwd_buffer[cwd_size]; // since only |getcwd| is standard, fix a buffer
            const char* working_directory_name=nullptr;
            const char* first_path = nullptr;
            #endif
            
            char lexical_break_chars[] = " \t\n=<>+-*/\\%,;:.()[]{}#!?$@\"|~";
 }

int yylex(YYSTYPE *valp, YYLTYPE *locp)
{ return atlas::interpreter::lex->get_token(valp,locp); }


void yyerror (YYLTYPE* locp, atlas::interpreter::expr_p* ,int* ,char const *s)
{ atlas::interpreter::main_input_buffer->show_range
  (std::cerr,
   locp->first_line, locp->first_column,
   locp->last_line,  locp->last_column);
  std::cerr << s << std::endl;
  atlas::interpreter::main_input_buffer->close_includes();
atlas::interpreter::clean=false;
}

extern "C" void sigint_handler (int)
{ atlas::interpreter::interrupt_flag=1; }

extern "C"
char* id_completion_func(const char* text, int state)
{ using namespace atlas::interpreter;
  static size_t l; static id_type i,n;
  if (state==0)
  { i=0; n=main_hash_table->nr_entries();
    l=std::strlen(text); // fix length for during search
  }
  while (i<n)
    // |i| is initialised above when |state==0|, and incremented below
  { id_type id=i++; // take next identifier code and increment
    const char* s=main_hash_table->name_of(id);
      // get stored identifier and increment loop
    if (std::strncmp(text,s,l) == 0 // is |text| a prefix of |s|?
      and (id<lex->first_identifier()  or
          not global_overload_table->variants(id).empty()  or
          global_id_table->present(id)))
    { char* result=static_cast<char*>(std::malloc(std::strlen(s)+1));
      if (result==nullptr)
        throw std::bad_alloc();
      return std::strcpy(result,s);
    }
  }
  return nullptr; /* if loop terminates, report failure */
}

#ifndef NREADLINE
extern "C" char** do_completion(const char* text, int start, int end)
{
  if (atlas::interpreter::lex->is_initial() and start>0)
   // could be a file name
  { int i; char c; bool need_file=false; bool in=true;
    for (i=0; i<start; ++i)
      if (std::isspace(c=rl_line_buffer[i]))
        continue; // ignore space characters
      else if (c=='<')
        need_file=true;
      else if (c=='>')
        need_file=true,in=false;
      else
        break; // any other preceding characters make it not a file name

    if (working_directory_name!=nullptr and need_file and i==start)
       // the sub-string is preceded by one or more copies of \.<, \.>
    { static_cast<void> // pretend we inspect the result, in fact ignore failure
      (chdir(in and first_path!=nullptr ? first_path : working_directory_name));
         // temporary \.{cd}
      return nullptr; // and signal that file name completion should be used
    }
  }
  rl_attempted_completion_over = true;
    // don't try file name completion if we get here
  return rl_completion_matches(text,id_completion_func);
    // but make |readline| call our function
}
#endif

namespace atlas { namespace interpreter {
                                         unsigned int input_path_size()
                                         { const row_value* path = force<row_value>(input_path_pointer->get());
                                         return path->val.size();
                                         }
                                         const std::string& input_path_component(unsigned int i)
                                         { const row_value* path = force<row_value>(input_path_pointer->get());
                                           const string_value* dir = force<string_value>(path->val[i].get());
                                         return dir->val;
                                         }

}}


int main(int argc, char** argv)
{ using namespace atlas::interpreter;


Hash_table hash; main_hash_table= &hash;
Id_table main_table;  global_id_table=&main_table;
overload_table main_overload_table;
  global_overload_table=&main_overload_table;

  bool use_readline=true;
  
  std::vector<const char*> paths,prelude_filenames;
  paths.reserve(argc-1); prelude_filenames.reserve(argc-1);
  
  while (*++argv!=nullptr)
  { static const char* const path_opt = "--path=";
    static const size_t pol = std::strlen(path_opt);
    std::string arg(*argv);
    if (arg=="--no-readline")
      { use_readline = false; continue; }
    if (arg.substr(0,pol)==path_opt)
       paths.push_back(&(*argv)[pol]);
    else prelude_filenames.push_back(*argv);
  }
BufferedInput input_buffer(isatty(STDIN_FILENO) ? "atlas> " : nullptr
                            ,use_readline ? readline : nullptr
			    ,use_readline ? add_history : nullptr);
  main_input_buffer= &input_buffer;
Lexical_analyser ana(input_buffer,hash,keywords,prim_names); lex=&ana;
  
  main_hash_table->match_literal("quiet");
  main_hash_table->match_literal("verbose");
  // these must be the very first identifiers
  ana.set_comment_delims('{','}');

  
  #ifndef NOT_UNIX
  working_directory_name = getcwd(cwd_buffer,cwd_size);
  #endif
  
  #ifndef NREADLINE
    rl_readline_name="axis"; // for users who want to customise |readline|
    using_history();
    rl_completer_word_break_characters = lexical_break_chars;
    rl_attempted_completion_function = do_completion; // set up input completion
  
  #endif
  signal(SIGINT,sigint_handler); // install handler for user interrupt
  initialise_evaluator();
  initialise_builtin_types();
  
  { 
    #ifndef NOT_UNIX
    if (paths.size()>0)
    { if (working_directory_name!=nullptr and chdir(paths[0])==0)
       // see if we can \.{cd} there
      { if (chdir(working_directory_name)!=0) // but if so switch back
          return EXIT_FAILURE; // in the unlikely case that we cannot, just give up
        first_path=paths[0]; // record that this is a valid directory
      }
      else
        std::cerr << "Warning: specified path '" << paths[0] 
                  << "' is not a directory." << std::endl;
    }
    #endif
    own_value input_path = std::make_shared<row_value>(paths.size());
    id_type ip_id = main_hash_table->match_literal("input_path");
    auto oit = force<row_value>(input_path.get())->val.begin();
    for (auto it=paths.begin(); it!=paths.end(); ++it)
      *oit++ = std::make_shared<string_value>(std::string(*it)+'/');
  global_id_table->add(ip_id
                         ,input_path, mk_type_expr("[string]"), false);
    input_path_pointer = global_id_table->address_of(ip_id);
  
    own_value prelude_log = std::make_shared<row_value>(0); // start out empty
    id_type pl_id = main_hash_table->match_literal("prelude_log");
    auto& logs = force<row_value>(input_path.get())->val;
    logs.reserve(prelude_filenames.size());
  global_id_table->add(pl_id
                         ,prelude_log, mk_type_expr("[string]"), true);
    prelude_log_pointer = global_id_table->address_of(pl_id);
  }

  
  for (auto it=prelude_filenames.begin(); it!=prelude_filenames.end(); ++it )
  { std::ostringstream log_stream; output_stream = &log_stream;
    main_input_buffer->push_file(*it,true);
      // set up to read |fname|, unless already done
    while (main_input_buffer->include_depth()>0) // go on until file ends
    { if (not ana.reset())
      { std::cerr << "Internal error, getline fails reading " << *it
                    << std::endl;
        return EXIT_FAILURE;
      }
      expr_p parse_tree;
      if (yyparse(&parse_tree,&verbosity)!=0)
        continue; // if a syntax error was signalled input has been closed
      if (verbosity!=0)
      { std::cerr << "Cannot "
                  << (verbosity<0 ? "quit" :
                      verbosity==1 ? "set verbose" : "redirect output")
                           << " during prelude.\n";
        verbosity=0; main_input_buffer->close_includes();
      }
      else
      { try
        { expression_ptr e; type_expr found_type=analyse_types(*parse_tree,e);
          e->evaluate(expression_base::single_value);
          if (found_type!=void_type)
            log_stream << "Value: " << *pop_value() << '\n';
          else
            pop_value(); // don't forget to cast away that void value
        }
        catch (std::exception& err)
        { std::cerr << err.what() << std::endl;
          reset_evaluator(); main_input_buffer->close_includes();
        clean=false;
        }
      }
      destroy_expr(parse_tree);
    }
    row_value* logs = force<row_value>(uniquify(*prelude_log_pointer));
    logs->val.emplace_back(std::make_shared<string_value>(log_stream.str()));
    output_stream = &std::cout;
  }
  std::cout << "This is 'atlas' (version " 
       << atlas::version::VERSION  << ", axis language version " 
       axis_version "),\n"  << atlas::version::NAME << 
       " interpreter,\ncompiled on "   << atlas::version::COMPILEDATE
       << ".   http://www.liegroups.org/\n";

  
  last_value = shared_value (new tuple_value(0));
  last_type = void_type.copy();
   // |last_type| is a |type_ptr| defined in \.{axis.w}
  while (ana.reset()) // get a fresh line for lexical analyser, or quit
  { 
    #ifndef NOT_UNIX
    if (working_directory_name!=nullptr)
      static_cast<void> // pretend we inspect the result, in fact ignore failure
          (chdir(working_directory_name)); // reset to true working directory
    #endif
    expr_p parse_tree;
    int old_verbosity=verbosity;
    std::ofstream redirect; // if opened, this will be closed at end of loop
    if (yyparse(&parse_tree,&verbosity)!=0)
       // syntax error (inputs are closed) or non-expression
      continue;
    if (verbosity!=0) // then some special action was requested
    { if (verbosity<0)
        break; // \.{quit} command
      if (verbosity==2 or verbosity==3)
        // indicates output redirection was requested
      { auto write_mode =
          verbosity==2 ? std::ios_base::trunc : std::ios_base::app;
        verbosity=old_verbosity; // ensure that verbosity change remains temporary
        
        { redirect.open(ana.scanned_file_name() ,std::ios_base::out | write_mode);
          if (redirect.is_open())
            output_stream = &redirect;
          else
          { std::cerr << "Failed to open " << ana.scanned_file_name() << std::endl;
            continue;
          }
        }
      }
      if (verbosity==1) //
        std::cout << "Expression before type analysis: " << *parse_tree
                  << std::endl;
    }
    
    { try
      { expression_ptr e;
        type_expr found_type=analyse_types(*parse_tree,e);
        if (verbosity>0)
          std::cout << "Type found: " << found_type << std::endl 
    	        << "Converted expression: " << *e << std::endl;
        if (found_type==void_type)
          e->void_eval();
        else
        { e->eval();
          last_type = std::move(found_type);
          last_value=pop_value();
          *output_stream << "Value: " << *last_value << std::endl;
        }
        destroy_expr(parse_tree);
      }
      
      catch (const runtime_error& err)
      { std::cerr << "Runtime error:\n  " << err.what() << "\nEvaluation aborted."
                  << std::endl;
      clean=false;
        reset_evaluator(); main_input_buffer->close_includes();
      }
      catch (const program_error& err)
      { std::cerr << err.what() << std::endl;
      clean=false;
        reset_evaluator(); main_input_buffer->close_includes();
      }
      catch (const logic_error& err)
      { std::cerr << "Internal error: " << err.what() << std::endl;
      clean=false;
        reset_evaluator(); main_input_buffer->close_includes();
      }
      catch (const std::exception& err)
      { std::cerr << err.what() << "\nEvaluation aborted.\n";
      clean=false;
        reset_evaluator(); main_input_buffer->close_includes();
      }
    }
    output_stream= &std::cout; // reset output stream if it was changed
  }
  signal(SIGINT,SIG_DFL); // reinstall default signal handler
  
  #ifndef NREADLINE
    clear_history();
     // clean up (presumably disposes of the lines stored in history)
  #endif
  std::cout << "Bye.\n";
  return clean ? EXIT_SUCCESS : EXIT_FAILURE ;
}


